from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Iterable, Iterator


class DataType(ABC):
    @staticmethod
    @abstractmethod
    def get_chunks(nb_workers: int, data: Any, **kwargs) -> Iterator[Any]:
        ...

    @staticmethod
    def get_work_extra(data: Any) -> Dict[str, Any]:
        return dict()

    @staticmethod
    @abstractmethod
    def work(
        data: Any,
        user_defined_function: Callable,
        user_defined_function_args: tuple,
        user_defined_function_kwargs: Dict[str, Any],
        extra: Dict[str, Any],
    ) -> Any:
        ...

    @staticmethod
    def get_reduce_extra(
        data: Any, user_defined_function_kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        return dict()

    @staticmethod
    @abstractmethod
    def reduce(datas: Iterable[Any], extra: Dict[str, Any]) -> Any:
        ...



from typing import Any, Callable, Dict, Iterable, Iterator
from types import GeneratorType

import pandas as pd

from ..utils import chunk, get_axis_int
from .generic import DataType


class DataFrame:
    class Apply(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, data: pd.DataFrame, **kwargs
        ) -> Iterator[pd.DataFrame]:
            user_defined_function_kwargs = kwargs["user_defined_function_kwargs"]

            axis_int = get_axis_int(user_defined_function_kwargs)
            opposite_axis_int = 1 - axis_int

            for chunk_ in chunk(data.shape[opposite_axis_int], nb_workers):
                yield data.iloc[chunk_] if axis_int == 1 else data.iloc[:, chunk_]

        @staticmethod
        def work(
            data: pd.DataFrame,
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> pd.DataFrame:
            return data.apply(
                user_defined_function,
                *user_defined_function_args,
                **user_defined_function_kwargs,
            )

        @staticmethod
        def get_reduce_extra(
            data: Any, user_defined_function_kwargs: Dict[str, Any]
        ) -> Dict[str, Any]:
            return {"axis": get_axis_int(user_defined_function_kwargs)}

        @staticmethod
        def reduce(
            datas: Iterable[pd.DataFrame], extra: Dict[str, Any]
        ) -> pd.DataFrame:
            if isinstance(datas, GeneratorType):
                datas = list(datas)
            axis = 0 if isinstance(datas[0], pd.Series) else 1 - extra["axis"]
            return pd.concat(datas, copy=False, axis=axis)

    class ApplyMap(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, data: pd.DataFrame, **kwargs
        ) -> Iterator[pd.DataFrame]:
            for chunk_ in chunk(data.shape[0], nb_workers):
                yield data.iloc[chunk_]

        @staticmethod
        def work(
            data: pd.DataFrame,
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> pd.DataFrame:
            return data.applymap(user_defined_function)

        @staticmethod
        def reduce(
            datas: Iterable[pd.DataFrame], extra: Dict[str, Any]
        ) -> pd.DataFrame:
            return pd.concat(datas, copy=False)



import itertools
from enum import Enum
from typing import Any, Dict, List, Tuple

import pandas as pd
from pandas import DataFrame, Index


def chunk(nb_item: int, nb_chunks: int, start_offset=0) -> List[slice]:
    """
    Return `nb_chunks` slices of approximatively `nb_item / nb_chunks` each.

    Parameters
    ----------
    nb_item : int
        Total number of items

    nb_chunks : int
        Number of chunks to return

    start_offset : int
        Shift start of slice by this amount

    Returns
    -------
    A list of slices

    Examples
    --------
    >>> chunks = chunk(103, 4)
    >>> chunks
    [slice(0, 26, None), slice(26, 52, None), slice(52, 78, None), slice(78, 103, None)]
    """
    if nb_item <= nb_chunks:
        return [slice(max(0, idx - start_offset), idx + 1) for idx in range(nb_item)]

    quotient = nb_item // nb_chunks
    remainder = nb_item % nb_chunks

    quotients = [quotient] * nb_chunks
    remainders = [1] * remainder + [0] * (nb_chunks - remainder)

    nb_elems_per_chunk = [
        quotient + remainder for quotient, remainder in zip(quotients, remainders)
    ]

    accumulated = list(itertools.accumulate(nb_elems_per_chunk))
    shifted_accumulated = accumulated.copy()
    shifted_accumulated.insert(0, 0)
    shifted_accumulated.pop()

    return [
        slice(max(0, begin - start_offset), end)
        for begin, end in zip(shifted_accumulated, accumulated)
    ]


def df_indexed_like(df: DataFrame, axes: List[Index]) -> bool:
    """
    Returns whether a data frame is indexed in the way specified by the
    provided axes.

    Used by DataFrameGroupBy to determine whether a group has been modified.

    Function adapted from pandas.core.groupby.ops._is_indexed_like

    Parameters
    ----------
    df : DataFrame
        The data frame in question

    axes : List[Index]
        The axes to which the data frame is compared

    Returns
    -------
    Whether or not the data frame is indexed in the same wa as the axes.
    """
    if isinstance(df, DataFrame):
        return df.axes[0].equals(axes[0])

    return False


def get_pandas_version() -> Tuple[int, int]:
    major_str, minor_str, *_ = pd.__version__.split(".")
    return int(major_str), int(minor_str)


def get_axis_int(user_defined_function_kwargs: Dict[str, Any]):
    axis = user_defined_function_kwargs.get("axis", 0)

    if axis not in {0, 1, "index", "columns"}:
        raise ValueError(f"No axis named {axis} for object type DataFrame")

    return {0: 0, 1: 1, "index": 0, "columns": 1}[axis]


class WorkerStatus(int, Enum):
    Running = 0
    Success = 1
    Error = 2



from typing import Any, Callable, Dict, Iterable, Iterator

import pandas as pd
from pandas.core.window.rolling import Rolling

from ..utils import chunk
from .generic import DataType


class SeriesRolling:
    class Apply(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, rolling: Rolling, **kwargs
        ) -> Iterator[pd.Series]:
            chunks = chunk(rolling.obj.size, nb_workers, rolling.window)

            for chunk_ in chunks:
                yield rolling.obj[chunk_]

        @staticmethod
        def get_work_extra(data: Rolling) -> Dict[str, Any]:
            return {
                "attributes": {
                    attribute: getattr(data, attribute)
                    for attribute in data._attributes
                }
            }

        @staticmethod
        def work(
            data: pd.Series,
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> pd.Series:
            attributes: Dict[str, Any] = extra["attributes"]
            worker_index: int = extra["worker_index"]

            result = data.rolling(**attributes).apply(
                user_defined_function,
                *user_defined_function_args,
                **user_defined_function_kwargs
            )

            return result if worker_index == 0 else result[attributes["window"] :]

        @staticmethod
        def reduce(datas: Iterable[pd.Series], extra: Dict[str, Any]) -> pd.Series:
            return pd.concat(datas, copy=False)



from .dataframe import DataFrame
from .dataframe_groupby import DataFrameGroupBy
from .expanding_groupby import ExpandingGroupBy
from .rolling_groupby import RollingGroupBy
from .generic import DataType
from .series import Series
from .series_rolling import SeriesRolling



import multiprocessing
from typing import Any, Callable, Dict, Iterable, Iterator, List, Tuple

import pandas as pd
from pandas.core.window.rolling import RollingGroupby as PandasRollingGroupby

from ..utils import WorkerStatus, chunk, get_pandas_version
from .generic import DataType


class RollingGroupBy:
    class Apply(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, data: PandasRollingGroupby, *args, **kwargs
        ) -> Iterator[List[Tuple[int, pd.DataFrame]]]:
            pandas_version = get_pandas_version()

            nb_items = (
                len(data._groupby) if pandas_version < (1, 3) else data._grouper.ngroups
            )

            chunks = chunk(nb_items, nb_workers)

            iterator = (
                iter(data._groupby)
                if pandas_version < (1, 3)
                else data._grouper.get_iterator(data.obj)
            )

            for chunk_ in chunks:
                yield [next(iterator) for _ in range(chunk_.stop - chunk_.start)]

        @staticmethod
        def get_work_extra(data: PandasRollingGroupby):
            attributes = {
                attribute: getattr(data, attribute) for attribute in data._attributes
            }

            return {"attributes": attributes}

        @staticmethod
        def work(
            data: List[Tuple[int, pd.DataFrame]],
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> List[pd.DataFrame]:
            show_progress_bars: bool = extra["show_progress_bars"]
            master_workers_queue: multiprocessing.Queue = extra["master_workers_queue"]
            worker_index: int = extra["worker_index"]

            def compute_result(
                iteration: int,
                attributes: Dict[str, Any],
                index: int,
                df: pd.DataFrame,
                user_defined_function: Callable,
                user_defined_function_args: tuple,
                user_defined_function_kwargs: Dict[str, Any],
            ) -> pd.DataFrame:
                item = df.rolling(**attributes).apply(
                    user_defined_function,
                    *user_defined_function_args,
                    **user_defined_function_kwargs
                )

                item.index = pd.MultiIndex.from_product([[index], item.index])

                if show_progress_bars:
                    master_workers_queue.put_nowait(
                        (worker_index, WorkerStatus.Running, iteration)
                    )

                return item

            attributes = extra["attributes"]
            attributes.pop("_grouper", None)

            dfs = (
                compute_result(
                    iteration,
                    attributes,
                    index,
                    df,
                    user_defined_function,
                    user_defined_function_args,
                    user_defined_function_kwargs,
                )
                for iteration, (index, df) in enumerate(data)
            )

            return pd.concat(dfs)

        @staticmethod
        def reduce(datas: Iterable[pd.DataFrame], extra: Dict[str, Any]) -> pd.Series:
            return pd.concat(datas, copy=False)



import math

import numpy as np
import pandas as pd
import pytest
from pandarallel import pandarallel


@pytest.fixture(params=(1000, 1))
def df_size(request):
    return request.param


@pytest.fixture(params=(False, True))
def progress_bar(request):
    return request.param


@pytest.fixture(params=(None, False))
def use_memory_fs(request):
    return request.param


@pytest.fixture(params=(RuntimeError, AttributeError, ZeroDivisionError))
def exception(request):
    return request.param


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_apply_axis_0(request):
    def func(x):
        return max(x) - min(x)

    return dict(named=func, anonymous=lambda x: max(x) - min(x))[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_apply_axis_1(request):
    def func(x):
        return math.sin(x.a**2) + math.sin(x.b**2)

    return dict(
        named=func, anonymous=lambda x: math.sin(x.a**2) + math.sin(x.b**2)
    )[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_applymap(request):
    def func(x):
        return math.sin(x**2) - math.cos(x**2)

    return dict(named=func, anonymous=lambda x: math.sin(x**2) - math.cos(x**2))[
        request.param
    ]


@pytest.fixture(params=("named", "anonymous"))
def func_series_map(request):
    def func(x):
        return math.log10(math.sqrt(math.exp(x**2)))

    return dict(
        named=func, anonymous=lambda x: math.log10(math.sqrt(math.exp(x**2)))
    )[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_series_apply(request):
    def func(x, power, bias=0):
        return math.log10(math.sqrt(math.exp(x**power))) + bias

    return dict(
        named=func,
        anonymous=lambda x, power, bias=0: math.log10(math.sqrt(math.exp(x**power)))
        + bias,
    )[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_series_rolling_apply(request):
    def func(x):
        return x.iloc[0] + x.iloc[1] ** 2 + x.iloc[2] ** 3 + x.iloc[3] ** 4

    return dict(
        named=func,
        anonymous=lambda x: x.iloc[0]
        + x.iloc[1] ** 2
        + x.iloc[2] ** 3
        + x.iloc[3] ** 4,
    )[request.param]


@pytest.fixture()
def func_dataframe_groupby_apply():
    def func(df):
        dum = 0
        for item in df.b:
            dum += math.log10(math.sqrt(math.exp(item**2)))

        return dum / len(df.b)

    return func


@pytest.fixture()
def func_dataframe_groupby_apply_complex():
    def func(df):
        return pd.DataFrame(
            [[df.b.mean(), df.b.min(), df.b.max()]],
            columns=["b_mean", "b_min", "b_max"],
        )

    return func


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_groupby_rolling_apply(request):
    def func(x):
        return x.iloc[0] + x.iloc[1] ** 2 + x.iloc[2] ** 3 + x.iloc[3] ** 4

    return dict(
        named=func,
        anonymous=lambda x: x.iloc[0]
        + x.iloc[1] ** 2
        + x.iloc[2] ** 3
        + x.iloc[3] ** 4,
    )[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_groupby_expanding_apply(request):
    def func(x):
        return (x.multiply(pd.Series(range(1, len(x)), dtype="float"))).sum()

    return dict(
        named=func,
        anonymous=lambda x: (
            x.multiply(pd.Series(range(1, len(x)), dtype="float"))
        ).sum(),
    )[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_apply_axis_0_no_reduce(request):
    def func(x):
        return x

    return dict(named=func, anonymous=lambda x: x)[request.param]


@pytest.fixture(params=("named", "anonymous"))
def func_dataframe_apply_axis_1_no_reduce(request):
    def func(x):
        return x**2

    return dict(named=func, anonymous=lambda x: x**2)[request.param]


@pytest.fixture
def pandarallel_init(progress_bar, use_memory_fs):
    pandarallel.initialize(
        progress_bar=progress_bar, use_memory_fs=use_memory_fs, nb_workers=2
    )


def test_dataframe_apply_invalid_function(pandarallel_init, exception):
    def f(_):
        raise exception

    df = pd.DataFrame(dict(a=[1, 2, 3, 4]))
    
    with pytest.raises(exception):
        df.parallel_apply(f)


def test_dataframe_apply_axis_0(pandarallel_init, func_dataframe_apply_axis_0, df_size):
    df = pd.DataFrame(
        dict(
            a=np.random.randint(1, 8, df_size),
            b=np.random.rand(df_size),
            c=np.random.randint(1, 8, df_size),
            d=np.random.rand(df_size),
            e=np.random.randint(1, 8, df_size),
            f=np.random.rand(df_size),
            g=np.random.randint(1, 8, df_size),
            h=np.random.rand(df_size),
        )
    )
    df.index = [item / 10 for item in df.index]

    res = df.apply(func_dataframe_apply_axis_0)
    res_parallel = df.parallel_apply(func_dataframe_apply_axis_0)
    assert res.equals(res_parallel)


def test_dataframe_apply_axis_1(pandarallel_init, func_dataframe_apply_axis_1, df_size):
    df = pd.DataFrame(
        dict(a=np.random.randint(1, 8, df_size), b=np.random.rand(df_size))
    )
    df.index = [item / 10 for item in df.index]

    res = df.apply(func_dataframe_apply_axis_1, axis=1)
    res_parallel = df.parallel_apply(func_dataframe_apply_axis_1, axis=1)
    assert res.equals(res_parallel)


def test_dataframe_apply_invalid_axis(pandarallel_init):
    df = pd.DataFrame(dict(a=[1, 2, 3, 4]))

    with pytest.raises(ValueError):
        df.parallel_apply(lambda x: x, axis="invalid")


def test_dataframe_applymap(pandarallel_init, func_dataframe_applymap, df_size):
    df = pd.DataFrame(
        dict(a=np.random.randint(1, 8, df_size), b=np.random.rand(df_size))
    )
    df.index = [item / 10 for item in df.index]

    res = df.applymap(func_dataframe_applymap)
    res_parallel = df.parallel_applymap(func_dataframe_applymap)
    assert res.equals(res_parallel)


def test_series_map(pandarallel_init, func_series_map, df_size):
    df = pd.DataFrame(dict(a=np.random.rand(df_size) + 1))

    res = df.a.map(func_series_map)
    res_parallel = df.a.parallel_map(func_series_map)
    assert res.equals(res_parallel)


def test_series_apply(pandarallel_init, func_series_apply, df_size):
    df = pd.DataFrame(dict(a=np.random.rand(df_size) + 1))

    res = df.a.apply(func_series_apply, args=(2,), bias=3)
    res_parallel = df.a.parallel_apply(func_series_apply, args=(2,), bias=3)
    assert res.equals(res_parallel)


def test_series_rolling_apply(pandarallel_init, func_series_rolling_apply, df_size):
    df = pd.DataFrame(dict(a=np.random.randint(1, 8, df_size), b=list(range(df_size))))

    res = df.b.rolling(4).apply(func_series_rolling_apply, raw=False)
    res_parallel = df.b.rolling(4).parallel_apply(func_series_rolling_apply, raw=False)

    assert res.equals(res_parallel)


def test_dataframe_groupby_apply(
    pandarallel_init, func_dataframe_groupby_apply, df_size
):
    df = pd.DataFrame(
        dict(
            a=np.random.randint(1, 8, df_size),
            b=np.random.rand(df_size),
            c=np.random.rand(df_size),
        )
    )

    res = df.groupby("a").apply(func_dataframe_groupby_apply)
    res_parallel = df.groupby("a").parallel_apply(func_dataframe_groupby_apply)
    assert res.equals(res_parallel)

    res = df.groupby(["a"]).apply(func_dataframe_groupby_apply)
    res_parallel = df.groupby(["a"]).parallel_apply(func_dataframe_groupby_apply)
    assert res.equals(res_parallel)

    res = df.groupby(["a", "b"]).apply(func_dataframe_groupby_apply)
    res_parallel = df.groupby(["a", "b"]).parallel_apply(func_dataframe_groupby_apply)
    assert res.equals(res_parallel)


def test_dataframe_groupby_apply_complex(
    pandarallel_init, func_dataframe_groupby_apply_complex, df_size
):
    df = pd.DataFrame(
        dict(a=np.random.randint(1, 100, df_size), b=np.random.rand(df_size))
    )

    res = df.groupby("a").apply(func_dataframe_groupby_apply_complex)
    res_parallel = df.groupby("a").parallel_apply(func_dataframe_groupby_apply_complex)
    res.equals(res_parallel)


def test_dataframe_groupby_rolling_apply(
    pandarallel_init, func_dataframe_groupby_rolling_apply, df_size
):
    df = pd.DataFrame(
        dict(a=np.random.randint(1, 10, df_size), b=np.random.rand(df_size))
    )

    res = (
        df.groupby("a")
        .b.rolling(4)
        .apply(func_dataframe_groupby_rolling_apply, raw=False)
    )
    res_parallel = (
        df.groupby("a")
        .b.rolling(4)
        .parallel_apply(func_dataframe_groupby_rolling_apply, raw=False)
    )
    assert res.equals(res_parallel)


def test_dataframe_groupby_expanding_apply(
    pandarallel_init, func_dataframe_groupby_expanding_apply, df_size
):
    df = pd.DataFrame(
        dict(a=np.random.randint(1, 10, df_size), b=np.random.rand(df_size))
    )

    res = (
        df.groupby("a")
        .b.expanding()
        .apply(func_dataframe_groupby_expanding_apply, raw=False)
    )
    res_parallel = (
        df.groupby("a")
        .b.expanding()
        .parallel_apply(func_dataframe_groupby_expanding_apply, raw=False)
    )
    res.equals(res_parallel)


def test_dataframe_axis_0_no_reduction(
    pandarallel_init, func_dataframe_apply_axis_0_no_reduce, df_size
):
    df = pd.DataFrame(
        dict(
            a=np.random.randint(1, 10, df_size),
            b=np.random.randint(1, 10, df_size),
            c=np.random.randint(1, 10, df_size),
        )
    )
    res = df.apply(func_dataframe_apply_axis_0_no_reduce)

    res_parallel = df.parallel_apply(func_dataframe_apply_axis_0_no_reduce)

    assert res.equals(res_parallel)


def test_dataframe_axis_1_no_reduction(
    pandarallel_init, func_dataframe_apply_axis_1_no_reduce, df_size
):
    df = pd.DataFrame(
        dict(
            a=np.random.randint(1, 10, df_size),
            b=np.random.randint(1, 10, df_size),
            c=np.random.randint(1, 10, df_size),
        )
    )

    res = df.apply(func_dataframe_apply_axis_1_no_reduce, axis=1)

    res_parallel = df.parallel_apply(func_dataframe_apply_axis_1_no_reduce, axis=1)

    assert res.equals(res_parallel)



import multiprocessing
import os
import shutil
import sys
from abc import ABC, abstractmethod
from enum import Enum
from itertools import count
from time import time_ns
from typing import Callable, List, Union

from .utils import WorkerStatus

INTERVAL_NS = 250_000_000  # 0.25 sec
MINIMUM_TERMINAL_WIDTH = 72


class ProgressBarsType(int, Enum):
    No = 0
    InUserDefinedFunction = 1
    InUserDefinedFunctionMultiplyByNumberOfColumns = 2
    InWorkFunction = 3


class ProgressBars(ABC):
    @abstractmethod
    def __init__(self, maxs: List[int], show: bool) -> None:
        ...

    @abstractmethod
    def update(self, values: List[int]) -> None:
        ...

    def set_error(self, index: int) -> None:
        pass


class ProgressState:
    def __init__(self, chunk_size: int) -> None:
        self.last_put_iteration = 0
        self.next_put_iteration = max(chunk_size // 100, 1)
        self.last_put_time = time_ns()


def is_notebook_lab() -> bool:
    try:
        shell: str = get_ipython().__class__.__name__  # type: ignore

        # Shell: Google Colab
        # TerminalInteractiveShell: Terminal running IPython
        # ZMQInteractiveShell: Jupyter notebook/lab or qtconsole
        return shell in {"Shell", "ZMQInteractiveShell"}
    except NameError:
        # Probably standard Python interpreter
        return False


class ProgressBarsConsole(ProgressBars):
    def __init__(self, maxs: List[int], show: bool) -> None:
        self.__show = show
        self.__bars = [[0, max] for max in maxs]
        self.__width = self.__get_width()

        self.__lines = self.__update_lines()

        if show:
            sys.stdout.write("\n".join(self.__lines))
            sys.stdout.flush()

    def __get_width(self) -> int:
        try:
            columns = shutil.get_terminal_size().columns
            return max(MINIMUM_TERMINAL_WIDTH, columns - 1)
        except AttributeError:
            # Python 2
            pass

        try:
            columns = int(os.popen("stty size", "r").read().split()[1])
            return max(MINIMUM_TERMINAL_WIDTH, columns - 1)
        except:
            return MINIMUM_TERMINAL_WIDTH

    def __remove_displayed_lines(self) -> None:
        if len(self.__bars) >= 1:
            sys.stdout.write("\b" * len(self.__lines[-1]))

        if len(self.__bars) >= 2:
            sys.stdout.write("\033M" * (len(self.__lines) - 1))

        self.__lines = []

    def __update_line(self, done: int, total: int) -> str:
        percent = done / total
        bar = (":" * int(percent * 40)).ljust(40, " ")
        percent = round(percent * 100, 2)
        format = " {percent:6.2f}% {bar:s} | {done:8d} / {total:8d} |"
        ret = format.format(percent=percent, bar=bar, done=done, total=total)
        return ret[: self.__width].ljust(self.__width, " ")

    def __update_lines(self) -> List[str]:
        return [self.__update_line(value, max) for value, max in self.__bars]

    def update(self, values: List[int]) -> None:
        """Update a bar value.
        Positional arguments:
        values - The new values of each bar
        """
        if not self.__show:
            return

        for index, value in enumerate(values):
            self.__bars[index][0] = value

        self.__remove_displayed_lines()
        self.__lines = self.__update_lines()

        sys.stdout.write("\n".join(self.__lines))
        sys.stdout.flush()


class ProgressBarsNotebookLab(ProgressBars):
    def __init__(self, maxs: List[int], show: bool) -> None:
        """Initialization.
        Positional argument:
        maxs - List containing the max value of each progress bar
        """
        self.__show = show

        if not show:
            return

        from IPython.display import display
        from ipywidgets import HBox, IntProgress, Label, VBox

        self.__bars = [
            HBox(
                [
                    IntProgress(0, 0, max, description="{:.2f}%".format(0)),
                    Label("{} / {}".format(0, max)),
                ]
            )
            for max in maxs
        ]

        display(VBox(self.__bars))

    def update(self, values: List[int]) -> None:
        """Update a bar value.
        Positional arguments:
        values - The new values of each bar
        """
        if not self.__show:
            return

        for index, value in enumerate(values):
            bar, label = self.__bars[index].children

            bar.value = value
            bar.description = "{:.2f}%".format(value / bar.max * 100)

            if value >= bar.max:
                bar.bar_style = "success"

            label.value = "{} / {}".format(value, bar.max)

    def set_error(self, index: int) -> None:
        """Set a bar on error"""
        if not self.__show:
            return

        bar, _ = self.__bars[index].children
        bar.bar_style = "danger"


def get_progress_bars(
    maxs: List[int], show
) -> Union[ProgressBarsNotebookLab, ProgressBarsConsole]:
    return (
        ProgressBarsNotebookLab(maxs, show)
        if is_notebook_lab()
        else ProgressBarsConsole(maxs, show)
    )


def progress_wrapper(
    user_defined_function: Callable,
    master_workers_queue: multiprocessing.Queue,
    index: int,
    chunk_size: int,
) -> Callable:
    """Wrap the function to apply in a function which monitor the part of work already
    done.
    """
    counter = count()
    state = ProgressState(chunk_size)

    def closure(*user_defined_function_args, **user_defined_functions_kwargs):
        iteration = next(counter)

        if iteration == state.next_put_iteration:
            time_now = time_ns()
            master_workers_queue.put_nowait((index, WorkerStatus.Running, iteration))

            delta_t = time_now - state.last_put_time
            delta_i = iteration - state.last_put_iteration

            state.next_put_iteration += (
                max(int((delta_i / delta_t) * INTERVAL_NS), 1) if delta_t != 0 else 1
            )

            state.last_put_iteration = iteration
            state.last_put_time = time_now

        return user_defined_function(
            *user_defined_function_args, **user_defined_functions_kwargs
        )

    return closure



import itertools
from typing import Any, Callable, Dict, Iterable, Iterator, List, Tuple, Union, cast

import pandas as pd
from pandas.core.groupby.generic import DataFrameGroupBy as PandasDataFrameGroupBy

from ..utils import chunk, df_indexed_like, get_pandas_version
from .generic import DataType


class DataFrameGroupBy:
    class Apply(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, dataframe_groupby: PandasDataFrameGroupBy, **kwargs
        ) -> Iterator[List[Tuple[int, pd.DataFrame]]]:
            chunks = chunk(dataframe_groupby.ngroups, nb_workers)
            iterator = iter(dataframe_groupby)

            for chunk_ in chunks:
                yield [next(iterator) for _ in range(chunk_.stop - chunk_.start)]

        @staticmethod
        def work(
            data: List[Tuple[int, pd.DataFrame]],
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> List[Tuple[int, pd.DataFrame, bool]]:
            def compute_result(
                key: int, df: pd.DataFrame
            ) -> Tuple[int, pd.DataFrame, bool]:
                result = user_defined_function(
                    df, *user_defined_function_args, **user_defined_function_kwargs
                )
                mutated = not df_indexed_like(result, df.axes)
                return key, result, mutated

            return [compute_result(key, df) for key, df in data]

        @staticmethod
        def get_reduce_extra(
            data: PandasDataFrameGroupBy, user_defined_function_kwargs: Dict[str, Any]
        ) -> Dict[str, Any]:
            return {"df_groupby": data}

        @staticmethod
        def reduce(
            datas: Iterable[List[Tuple[int, pd.DataFrame, bool]]], extra: Dict[str, Any]
        ) -> pd.Series:
            def get_args(
                keys: List[int],
                values: List[pd.DataFrame],
                df_groupby: PandasDataFrameGroupBy,
            ) -> Union[
                Tuple[List[int], List[pd.DataFrame]],
                Tuple[pd.DataFrame, List[int], List[pd.DataFrame]],
                Tuple[pd.DataFrame, List[pd.DataFrame]],
            ]:
                pandas_version = get_pandas_version()

                if pandas_version < (1, 3):
                    return keys, values
                elif pandas_version < (1, 4):
                    return df_groupby._selected_obj, keys, values
                else:
                    return df_groupby._selected_obj, values

            df_groupby: PandasDataFrameGroupBy = extra["df_groupby"]

            results = itertools.chain.from_iterable(datas)
            keys, values, mutated = zip(*results)

            keys = cast(List[int], keys)
            values = cast(List[pd.DataFrame], values)
            mutated = cast(List[bool], mutated)

            args = get_args(keys, values, df_groupby)

            return df_groupby._wrap_applied_output(
                *args, not_indexed_same=df_groupby.mutated or mutated
            )



from typing import Any, Callable, Dict, Iterable, Iterator

import pandas as pd

from ..utils import chunk
from .generic import DataType


class Series:
    class Apply(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, data: pd.Series, **kwargs
        ) -> Iterator[pd.Series]:
            for chunk_ in chunk(data.size, nb_workers):
                yield data.iloc[chunk_]

        @staticmethod
        def work(
            data: pd.Series,
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> pd.Series:
            return data.apply(
                user_defined_function,
                *user_defined_function_args,
                **user_defined_function_kwargs
            )

        @staticmethod
        def reduce(datas: Iterable[pd.Series], extra: Dict[str, Any]) -> pd.Series:
            return pd.concat(datas, copy=False)

    class Map(DataType):
        @staticmethod
        def get_chunks(
            nb_workers: int, data: pd.Series, **kwargs
        ) -> Iterator[pd.Series]:
            for chunk_ in chunk(data.size, nb_workers):
                yield data.iloc[chunk_]

        @staticmethod
        def work(
            data: pd.Series,
            user_defined_function: Callable,
            user_defined_function_args: tuple,
            user_defined_function_kwargs: Dict[str, Any],
            extra: Dict[str, Any],
        ) -> pd.Series:
            return data.map(
                user_defined_function,
                *user_defined_function_args,
                **user_defined_function_kwargs
            )

        @staticmethod
        def reduce(datas: Iterable[pd.Series], extra: Dict[str, Any]) -> pd.Series:
            return pd.concat(datas, copy=False)



import multiprocessing
import os
import pickle
from itertools import count
from multiprocessing.managers import SyncManager
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, Iterator, Optional, Tuple, Type, cast

import dill
import pandas as pd
import psutil
from pandas.core.groupby import DataFrameGroupBy as PandaDataFrameGroupBy
from pandas.core.window.expanding import ExpandingGroupby as PandasExpandingGroupby
from pandas.core.window.rolling import RollingGroupby as PandasRollingGroupby

from .data_types import (
    DataFrame,
    DataFrameGroupBy,
    DataType,
    ExpandingGroupBy,
    RollingGroupBy,
    Series,
    SeriesRolling,
)
from .progress_bars import ProgressBarsType, get_progress_bars, progress_wrapper
from .utils import WorkerStatus

ON_WINDOWS = os.name == "nt"
CONTEXT = multiprocessing.get_context("spawn" if ON_WINDOWS else "fork")

# Root of Memory File System
MEMORY_FS_ROOT = "/dev/shm"

# By default, Pandarallel use all available CPUs
NB_PHYSICAL_CORES = psutil.cpu_count(logical=False)

# Prefix and suffix for files used with Memory File System
PREFIX = "pandarallel"
PREFIX_INPUT = f"{PREFIX}_input_"
PREFIX_OUTPUT = f"{PREFIX}_output_"
SUFFIX = ".pickle"

# We use these classes decorators pattern instead of the classic one because of this:
# https://www.stevenengelhardt.com/2013/01/16/python-multiprocessing-module-and-closures/


class WrapWorkFunctionForFileSystem:
    def __init__(
        self,
        work_function: Callable[
            [Any, Callable, tuple, Dict[str, Any], Dict[str, Any]], Any
        ],
    ) -> None:
        self.work_function = work_function

    def __call__(
        self,
        input_file_path: Path,
        output_file_path: Path,
        progress_bars_type: ProgressBarsType,
        worker_index: int,
        master_workers_queue: multiprocessing.Queue,
        dilled_user_defined_function: bytes,
        user_defined_function_args: tuple,
        user_defined_function_kwargs: Dict[str, Any],
        extra: Dict[str, Any],
    ) -> None:
        try:
            # Load dataframe from input file
            with input_file_path.open("rb") as file_descriptor:
                data = pickle.load(file_descriptor)

            # Delete input file since we don't need it any more. It will free some RAM
            # since the input file is stored into Shared Memory.
            input_file_path.unlink()

            data_size = len(data)
            user_defined_function: Callable = dill.loads(dilled_user_defined_function)

            progress_wrapped_user_defined_function = progress_wrapper(
                user_defined_function, master_workers_queue, worker_index, data_size
            )

            used_user_defined_function = (
                progress_wrapped_user_defined_function
                if progress_bars_type
                in (
                    ProgressBarsType.InUserDefinedFunction,
                    ProgressBarsType.InUserDefinedFunctionMultiplyByNumberOfColumns,
                )
                else user_defined_function
            )

            result = self.work_function(
                data,
                used_user_defined_function,
                user_defined_function_args,
                user_defined_function_kwargs,
                extra,
            )

            with output_file_path.open("wb") as file_descriptor:
                pickle.dump(result, file_descriptor)

            master_workers_queue.put((worker_index, WorkerStatus.Success, None))

        except:
            master_workers_queue.put((worker_index, WorkerStatus.Error, None))
            raise


class WrapWorkFunctionForPipe:
    def __init__(
        self,
        work_function: Callable[
            [
                Any,
                Callable,
                tuple,
                Dict[str, Any],
                Dict[str, Any],
            ],
            Any,
        ],
    ) -> None:
        self.work_function = work_function

    def __call__(
        self,
        data: Any,
        progress_bars_type: ProgressBarsType,
        worker_index: int,
        master_workers_queue: multiprocessing.Queue,
        dilled_user_defined_function: bytes,
        user_defined_function_args: tuple,
        user_defined_function_kwargs: Dict[str, Any],
        extra: Dict[str, Any],
    ) -> Any:
        try:
            data_size = len(data)
            user_defined_function: Callable = dill.loads(dilled_user_defined_function)

            progress_wrapped_user_defined_function = progress_wrapper(
                user_defined_function, master_workers_queue, worker_index, data_size
            )

            used_user_defined_function = (
                progress_wrapped_user_defined_function
                if progress_bars_type
                in (
                    ProgressBarsType.InUserDefinedFunction,
                    ProgressBarsType.InUserDefinedFunctionMultiplyByNumberOfColumns,
                )
                else user_defined_function
            )

            results = self.work_function(
                data,
                used_user_defined_function,
                user_defined_function_args,
                user_defined_function_kwargs,
                extra,
            )

            master_workers_queue.put((worker_index, WorkerStatus.Success, None))

            return results

        except:
            master_workers_queue.put((worker_index, WorkerStatus.Error, None))
            raise


def wrap_reduce_function_for_file_system(
    reduce_function: Callable[[Iterator, Dict[str, Any]], Any]
) -> Callable[[Iterator[Path], Dict[str, Any]], Any]:
    """This wrapper transforms a `reduce` function which takes as input:
    - A list of pandas Dataframe
    - An user defined function
    and which returns a pandas Dataframe, into a `reduct` function which takes as input:
    - A list of paths where  pandas Dataframe are pickled
    which returns a pandas Dataframe.
    """

    def closure(output_file_paths: Iterator[Path], extra: Dict[str, Any]) -> Any:
        def get_dataframe_and_delete_file(file_path: Path) -> Any:
            with file_path.open("rb") as file_descriptor:
                data = pickle.load(file_descriptor)

            file_path.unlink()
            return data

        dfs = (
            get_dataframe_and_delete_file(output_file_path)
            for output_file_path in output_file_paths
        )

        return reduce_function(dfs, extra)

    return closure


def parallelize_with_memory_file_system(
    nb_requested_workers: int,
    data_type: Type[DataType],
    progress_bars_type: ProgressBarsType,
):
    def closure(
        data: Any,
        user_defined_function: Callable,
        *user_defined_function_args: tuple,
        **user_defined_function_kwargs: Dict[str, Any],
    ):
        wrapped_work_function = WrapWorkFunctionForFileSystem(data_type.work)
        wrapped_reduce_function = wrap_reduce_function_for_file_system(data_type.reduce)

        chunks = list(
            data_type.get_chunks(
                nb_requested_workers,
                data,
                user_defined_function_kwargs=user_defined_function_kwargs,
            )
        )

        nb_workers = len(chunks)

        multiplicator_factor = (
            len(cast(pd.DataFrame, data).columns)
            if progress_bars_type
            == ProgressBarsType.InUserDefinedFunctionMultiplyByNumberOfColumns
            else 1
        )

        progresses_length = [len(chunk_) * multiplicator_factor for chunk_ in chunks]

        work_extra = data_type.get_work_extra(data)
        reduce_extra = data_type.get_reduce_extra(data, user_defined_function_kwargs)

        show_progress_bars = progress_bars_type != ProgressBarsType.No

        progress_bars = get_progress_bars(progresses_length, show_progress_bars)
        progresses = [0] * nb_workers
        workers_status = [WorkerStatus.Running] * nb_workers

        input_files = [
            NamedTemporaryFile(
                prefix=PREFIX_INPUT, suffix=SUFFIX, dir=MEMORY_FS_ROOT, delete=False
            )
            for _ in range(nb_workers)
        ]

        output_files = [
            NamedTemporaryFile(
                prefix=PREFIX_OUTPUT, suffix=SUFFIX, dir=MEMORY_FS_ROOT, delete=False
            )
            for _ in range(nb_workers)
        ]

        try:
            for chunk, input_file in zip(chunks, input_files):
                with Path(input_file.name).open("wb") as file_descriptor:
                    pickle.dump(chunk, file_descriptor)

            dilled_user_defined_function = dill.dumps(user_defined_function)
            manager: SyncManager = CONTEXT.Manager()
            master_workers_queue = manager.Queue()

            work_args_list = [
                (
                    Path(input_file.name),
                    Path(output_file.name),
                    progress_bars_type,
                    worker_index,
                    master_workers_queue,
                    dilled_user_defined_function,
                    user_defined_function_args,
                    user_defined_function_kwargs,
                    {
                        **work_extra,
                        **{
                            "master_workers_queue": master_workers_queue,
                            "show_progress_bars": show_progress_bars,
                            "worker_index": worker_index,
                        },
                    },
                )
                for worker_index, (
                    input_file,
                    output_file,
                ) in enumerate(zip(input_files, output_files))
            ]

            pool = CONTEXT.Pool(nb_workers)
            results_promise = pool.starmap_async(wrapped_work_function, work_args_list)

            pool.close()

            generation = count()

            while any(
                (
                    worker_status == WorkerStatus.Running
                    for worker_status in workers_status
                )
            ):
                message: Tuple[int, WorkerStatus, Any] = master_workers_queue.get()
                worker_index, worker_status, payload = message
                workers_status[worker_index] = worker_status

                if worker_status == WorkerStatus.Success:
                    progresses[worker_index] = progresses_length[worker_index]
                    progress_bars.update(progresses)
                elif worker_status == WorkerStatus.Running:
                    progress = cast(int, payload)
                    progresses[worker_index] = progress

                    if next(generation) % nb_workers == 0:
                        progress_bars.update(progresses)
                elif worker_status == WorkerStatus.Error:
                    progress_bars.set_error(worker_index)
                    progress_bars.update(progresses)

            try:
                return wrapped_reduce_function(
                    (Path(output_file.name) for output_file in output_files),
                    reduce_extra,
                )
            except EOFError:
                # Loading the files failed, this most likely means that there
                # was some error during processing and the files were never
                # saved at all. 
                results_promise.get()

                # If the above statement does not raise an exception, that
                # means the multiprocessing went well and we want to re-raise
                # the original EOFError.
                raise

        finally:
            for output_file in output_files:
                # When pandarallel stop supporting Python 3.7 and older, replace this
                # try/except clause by:
                # Path(output_file.name).unlink(missing_ok=True)
                try:
                    Path(output_file.name).unlink()
                except FileNotFoundError:
                    # Do nothing, this is the nominal case.
                    pass

    return closure


def parallelize_with_pipe(
    nb_requested_workers: int,
    data_type: Type[DataType],
    progress_bars_type: ProgressBarsType,
):
    def closure(
        data: Any,
        user_defined_function: Callable,
        *user_defined_function_args: tuple,
        **user_defined_function_kwargs: Dict[str, Any],
    ):
        wrapped_work_function = WrapWorkFunctionForPipe(data_type.work)
        dilled_user_defined_function = dill.dumps(user_defined_function)
        manager: SyncManager = CONTEXT.Manager()
        master_workers_queue = manager.Queue()

        chunks = list(
            data_type.get_chunks(
                nb_requested_workers,
                data,
                user_defined_function_kwargs=user_defined_function_kwargs,
            )
        )

        nb_workers = len(chunks)

        multiplicator_factor = (
            len(cast(pd.DataFrame, data).columns)
            if progress_bars_type
            == ProgressBarsType.InUserDefinedFunctionMultiplyByNumberOfColumns
            else 1
        )

        progresses_length = [len(chunk_) * multiplicator_factor for chunk_ in chunks]

        work_extra = data_type.get_work_extra(data)
        reduce_extra = data_type.get_reduce_extra(data, user_defined_function_kwargs)

        show_progress_bars = progress_bars_type != ProgressBarsType.No

        progress_bars = get_progress_bars(progresses_length, show_progress_bars)
        progresses = [0] * nb_workers
        workers_status = [WorkerStatus.Running] * nb_workers

        work_args_list = [
            (
                chunk,
                progress_bars_type,
                worker_index,
                master_workers_queue,
                dilled_user_defined_function,
                user_defined_function_args,
                user_defined_function_kwargs,
                {
                    **work_extra,
                    **{
                        "master_workers_queue": master_workers_queue,
                        "show_progress_bars": show_progress_bars,
                        "worker_index": worker_index,
                    },
                },
            )
            for worker_index, chunk in enumerate(chunks)
        ]

        pool = CONTEXT.Pool(nb_workers)
        results_promise = pool.starmap_async(wrapped_work_function, work_args_list)
        pool.close()

        generation = count()

        while any(
            (worker_status == WorkerStatus.Running for worker_status in workers_status)
        ):
            message: Tuple[int, WorkerStatus, Any] = master_workers_queue.get()
            worker_index, worker_status, payload = message
            workers_status[worker_index] = worker_status

            if worker_status == WorkerStatus.Success:
                progresses[worker_index] = progresses_length[worker_index]
                progress_bars.update(progresses)
            elif worker_status == WorkerStatus.Running:
                progress = cast(int, payload)
                progresses[worker_index] = progress

                if next(generation) % nb_workers == 0:
                    progress_bars.update(progresses)
            elif worker_status == WorkerStatus.Error:
                progress_bars.set_error(worker_index)

        results = results_promise.get()

        return data_type.reduce(results, reduce_extra)

    return closure


class pandarallel:
    @classmethod
    def initialize(
        cls,
        shm_size_mb=None,
        nb_workers=NB_PHYSICAL_CORES,
        progress_bar=False,
        verbose=2,
        use_memory_fs: Optional[bool] = None,
    ) -> None:
        show_progress_bars = progress_bar
        is_memory_fs_available = Path(MEMORY_FS_ROOT).exists()

        use_memory_fs = (
            use_memory_fs if use_memory_fs is not None else is_memory_fs_available
        )

        parallelize = (
            parallelize_with_memory_file_system
            if use_memory_fs
            else parallelize_with_pipe
        )

        if use_memory_fs and not is_memory_fs_available:
            raise SystemError("Memory file system is not available")

        if verbose >= 2:
            print(f"INFO: Pandarallel will run on {nb_workers} workers.")

            message = (
                (
                    "INFO: Pandarallel will use Memory file system to transfer data "
                    "between the main process and workers."
                )
                if use_memory_fs
                else (
                    "INFO: Pandarallel will use standard multiprocessing data transfer "
                    "(pipe) to transfer data between the main process and workers."
                )
            )

            print(message)

            if ON_WINDOWS and verbose >= 2:
                print()
                print(
                    (
                        "WARNING: You are on Windows. If you detect any issue with "
                        "pandarallel, be sure you checked out the Troubleshooting page:"
                    )
                )
                print("https://nalepae.github.io/pandarallel/troubleshooting/")

        progress_bars_in_user_defined_function = (
            ProgressBarsType.InUserDefinedFunction
            if show_progress_bars
            else ProgressBarsType.No
        )

        progress_bars_in_user_defined_function_multiply_by_number_of_columns = (
            ProgressBarsType.InUserDefinedFunctionMultiplyByNumberOfColumns
            if show_progress_bars
            else ProgressBarsType.No
        )

        progress_bars_in_work_function = (
            ProgressBarsType.InWorkFunction
            if show_progress_bars
            else ProgressBarsType.No
        )

        # DataFrame
        pd.DataFrame.parallel_apply = parallelize(
            nb_workers, DataFrame.Apply, progress_bars_in_user_defined_function
        )
        pd.DataFrame.parallel_applymap = parallelize(
            nb_workers,
            DataFrame.ApplyMap,
            progress_bars_in_user_defined_function_multiply_by_number_of_columns,
        )

        # DataFrame GroupBy
        PandaDataFrameGroupBy.parallel_apply = parallelize(
            nb_workers, DataFrameGroupBy.Apply, progress_bars_in_user_defined_function
        )

        # Expanding GroupBy
        PandasExpandingGroupby.parallel_apply = parallelize(
            nb_workers, ExpandingGroupBy.Apply, progress_bars_in_work_function
        )

        # Rolling GroupBy
        PandasRollingGroupby.parallel_apply = parallelize(
            nb_workers, RollingGroupBy.Apply, progress_bars_in_work_function
        )

        # Series
        pd.Series.parallel_apply = parallelize(
            nb_workers, Series.Apply, progress_bars_in_user_defined_function
        )
        pd.Series.parallel_map = parallelize(nb_workers, Series.Map, show_progress_bars)

        # Series Rolling
        pd.core.window.Rolling.parallel_apply = parallelize(
            nb_workers, SeriesRolling.Apply, progress_bars_in_user_defined_function
        )
